#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	named plugins
# revision: 3		author: OrangeMohawk
# 	Standardized Event Reflexives
# revision: 4		author: Lord Zedd
# 	Updating
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

mulg_runtime_assault_event_event = (
    "game_start",
    "bomb_taken",
    "bomb_dropped",
    "bomb_returned_by_player",
    "bomb_returned_by_timeout",
    "bomb_captured",
    "bomb_new_defensive_team",
    "bomb_return_failure",
    "side_switch_tick",
    "side_switch_final_tick",
    "side_switch_30_seconds",
    "side_switch_10_seconds",
    "bomb_returned_by_defusing",
    "bomb_placed_on_enemy_post",
    "bomb_arming_started",
    "bomb_arming_completed",
    "bomb_contested",
    "bomb_returned_neutral",
    )

mulg_runtime_ctf_event_event = (
    "game_start",
    "flag_taken",
    "flag_taken_2",
    "flag_dropped",
    "flag_returned_by_player",
    "flag_returned_by_timeout",
    "flag_captured",
    "flag_new_defensive_team",
    "flag_return_failure",
    "side_switch_tick",
    "side_switch_final_tick",
    "side_switch_30_seconds",
    "side_switch_10_seconds",
    "flag_contested",
    "flag_capture_failure",
    )

mulg_runtime_flavor_event_event = (
    "extermination",
    "perfection",
    "double_kill",
    "triple_kill",
    "overkill",
    "killtacular",
    "killtrocity",
    "killimanjaro",
    "killtastrophe",
    "killpocalypse",
    "killionaire",
    "killing_spree",
    "killing_frenzy",
    "running_riot",
    "rampage",
    "untouchable",
    "invincible",
    "sniper_spree",
    "sharpshooter",
    "shotgun_spree",
    "open_season",
    "splatter_spree",
    "vehicular_manslaughter",
    "sword_spree",
    "slice_n_dice",
    "juggernaut_spree",
    "unstoppable",
    "infection_spree",
    "mmmm_brains",
    "zombie_killing_spree",
    "hell_s_janitor",
    "is_quisnam_protero_damno",
    "hail_to_the_king",
    "bulltrue",
    "splatter",
    "highjack",
    "skyjack",
    "death_from_the_grave",
    "killjoy",
    "laser_kill",
    "sticky_kill",
    "sniper_kill",
    "assassin",
    "beat_down",
    "incineration",
    "wheelman",
    "bomb_planted",
    "killed_bomb_carrier",
    "killed_vip",
    "killed_juggernaut",
    "oddball_kill",
    "flag_score",
    "killed_flag_carrier",
    "flag_kill",
    "last_man_standing",
    "sandbox_not_enough_room",
    "sandbox_too_many_on_map",
    )

mulg_runtime_general_event_audience = (
    "cause_player",
    "cause_team",
    "effect_player",
    "effect_team",
    "all",
    )

mulg_runtime_general_event_event = (
    "kill",
    "suicide",
    "betrayal",
    "victory",
    "team_victory",
    "unused1",
    "unused2",
    "_1_min_to_win",
    "team_1_min_to_win",
    "_30_secs_to_win",
    "team_30_secs_to_win",
    "player_quit",
    "player_joined",
    "killed_by_unknown",
    "_30_minutes_left",
    "_15_minutes_left",
    "_5_minutes_left",
    "_1_minute_left",
    "time_expired",
    "game_over",
    "respawn_tick",
    "last_respawn_tick",
    "teleporter_used",
    "teleporter_blocked",
    "player_changed_team",
    "player_rejoined",
    "gained_lead",
    "gained_team_lead",
    "lost_lead",
    "lost_team_lead",
    "tied_leader",
    "tied_team_leader",
    "round_over",
    "_30_seconds_left",
    "_10_seconds_left",
    "kill_falling",
    "kill_collision",
    "kill_melee",
    "sudden_death",
    "player_booted_player",
    "kill_flag_carrier",
    "kill_bomb_carrier",
    "kill_sticky_grenade",
    "kill_sniper",
    "kill_standard_melee",
    "boarded_vehicle",
    "start_team_notice",
    "telefrag",
    "_10_secs_to_win",
    "team_10_secs_to_win",
    "kill_bulltrue",
    "kill_post_mortem",
    "highjack",
    "skyjack",
    "kill_laser",
    "kill_fire",
    "wheelman",
    )

mulg_runtime_general_event_required_field = (
    "none",
    "cause_player",
    "cause_team",
    "effect_player",
    "effect_team",
    )

mulg_runtime_general_event_team = (
    "none_player_only",
    "cause",
    "effect",
    "all",
    )

mulg_runtime_general_event_type = (
    "general",
    "flavor",
    "slayer",
    "capture_the_flag",
    "oddball",
    "unused",
    "king_of_the_hill",
    "vip",
    "juggernaut",
    "territories",
    "assault",
    "infection",
    )

mulg_runtime_infection_event_event = (
    "game_start",
    "infected",
    "new_zombie",
    "new_alpha_zombie",
    "last_man_standing",
    )

mulg_runtime_juggernaut_event_event = (
    "game_start",
    "new_juggernaut",
    "destination_move",
    "juggernaut_killed",
    )

mulg_runtime_king_of_the_hill_event_event = (
    "game_start",
    "hill_controlled",
    "hill_contested",
    "hill_tick",
    "hill_move",
    "hill_controlled_team",
    "hill_contested_team",
    )

mulg_runtime_oddball_event_event = (
    "game_start",
    "ball_spawned",
    "ball_picked_up",
    "ball_dropped",
    "ball_reset",
    "ball_tick",
    "_10_points_remaining",
    "_25_points_remaining",
    )

mulg_runtime_slayer_event_event = (
    "game_start",
    "new_target",
    )

mulg_runtime_state_response_state = (
    "waiting_for_space_to_clear",
    "observing",
    "respawning_soon",
    "sitting_out",
    "out_of_lives",
    "playing_winning",
    "playing_tied",
    "playing_losing",
    "game_over_won",
    "game_over_tied",
    "game_over_lost",
    "game_over_tied2",
    "you_have_flag",
    "enemy_has_flag",
    "flag_not_home",
    "carrying_oddball",
    "you_are_juggernaut",
    "you_control_hill",
    "switching_sides_soon",
    "player_recently_started",
    "you_have_bomb",
    "flag_contested",
    "bomb_contested",
    "limited_lives_multiple",
    "limited_lives_single",
    "limited_lives_final",
    "playing_winning_unlimited",
    "playing_tied_unlimited",
    "playing_losing_unlimited",
    )

mulg_runtime_territories_event_event = (
    "game_start",
    "territory_control_gained",
    "territory_contest_lost",
    "all_territories_controlled",
    "new_defensive_team",
    )

mulg_runtime_vip_event_event = (
    "game_start",
    "new_vip",
    "killed_vip",
    "vip_death",
    "vip_notice",
    "arrived_at_destination",
    "destination_move",
    )


mulg_universal_team_color = Struct("team_color",
    color_rgb_float("color"),
    ENDIAN=">", SIZE=12
    )


mulg_universal_armor_customization_region_permuation_variant = Struct("variant",
    h3_string_id("region"),
    h3_string_id("permutation"),
    ENDIAN=">", SIZE=8
    )


mulg_universal_armor_customization_region_permuation = Struct("permuation",
    h3_string_id("name"),
    h3_string_id("description"),
    Bool16("flags",
        "has_requirement",
        "has_special_requirement",
        ),
    SInt16("unknown"),
    h3_string_id("achievement_requirement"),
    h3_reflexive("variant", mulg_universal_armor_customization_region_permuation_variant),
    ENDIAN=">", SIZE=28
    )


mulg_universal_armor_customization_region = Struct("region",
    h3_string_id("name"),
    h3_reflexive("permuations", mulg_universal_armor_customization_region_permuation),
    ENDIAN=">", SIZE=16
    )


mulg_universal_armor_customization = Struct("armor_customization",
    h3_string_id("character_name"),
    h3_reflexive("regions", mulg_universal_armor_customization_region),
    ENDIAN=">", SIZE=16
    )


mulg_universal_game_variant_weapon = Struct("game_variant_weapon",
    h3_string_id("name"),
    Float("random_chance"),
    h3_dependency("weapon"),
    ENDIAN=">", SIZE=24
    )


mulg_universal_game_variant_vehicle = Struct("game_variant_vehicle",
    h3_string_id("name"),
    h3_dependency("vehicle"),
    ENDIAN=">", SIZE=20
    )


mulg_universal_game_variant_equipment = Struct("game_variant_equipment",
    h3_string_id("name"),
    h3_dependency("grenade"),
    ENDIAN=">", SIZE=20
    )


mulg_universal_weapon_set_substitution = Struct("substitution",
    h3_string_id("original_weapon"),
    h3_string_id("substituted_weapon"),
    ENDIAN=">", SIZE=8
    )


mulg_universal_weapon_set = Struct("weapon_set",
    h3_string_id("name"),
    h3_reflexive("substitutions", mulg_universal_weapon_set_substitution),
    ENDIAN=">", SIZE=16
    )


mulg_universal_vehicle_set_substitution = Struct("substitution",
    h3_string_id("original_vehicle"),
    h3_string_id("substituted_vehicle"),
    ENDIAN=">", SIZE=8
    )


mulg_universal_vehicle_set = Struct("vehicle_set",
    h3_string_id("name"),
    h3_reflexive("substitutions", mulg_universal_vehicle_set_substitution),
    ENDIAN=">", SIZE=16
    )


mulg_universal = Struct("universal",
    h3_dependency("random_player_name_strings"),
    h3_dependency("team_name_strings"),
    h3_reflexive("team_colors", mulg_universal_team_color),
    h3_reflexive("armor_customization", mulg_universal_armor_customization),
    h3_dependency("multiplayer_strings"),
    h3_dependency("sandbox_ui_strings"),
    h3_dependency("sandbox_ui_properties"),
    h3_reflexive("game_variant_weapons", mulg_universal_game_variant_weapon),
    h3_reflexive("game_variant_vehicles", mulg_universal_game_variant_vehicle),
    h3_reflexive("game_variant_equipment", mulg_universal_game_variant_equipment),
    h3_reflexive("weapon_sets", mulg_universal_weapon_set),
    h3_reflexive("vehicle_sets", mulg_universal_vehicle_set),
    h3_dependency("engine_settings"),
    ENDIAN=">", SIZE=180
    )


mulg_runtime_sound = Struct("sound",
    h3_dependency("sound"),
    ENDIAN=">", SIZE=16
    )


mulg_runtime_looping_sound = Struct("looping_sound",
    h3_dependency("looping_sound"),
    ENDIAN=">", SIZE=16
    )


mulg_runtime_general_event = Struct("general_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_general_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_flavor_event = Struct("flavor_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_flavor_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_slayer_event = Struct("slayer_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_slayer_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_ctf_event = Struct("ctf_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_ctf_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_oddball_event = Struct("oddball_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_oddball_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_king_of_the_hill_event = Struct("king_of_the_hill_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_king_of_the_hill_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_vip_event = Struct("vip_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_vip_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_juggernaut_event = Struct("juggernaut_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_juggernaut_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_territories_event = Struct("territories_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_territories_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_assault_event = Struct("assault_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_assault_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_infection_event = Struct("infection_event",
    Bool16("flags", *unknown_flags_16),
    SEnum16("type", *mulg_runtime_general_event_type),
    SEnum16("event", *mulg_runtime_infection_event_event),
    SEnum16("audience", *mulg_runtime_general_event_audience),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    SEnum16("team", *mulg_runtime_general_event_team),
    SInt16("unknown_2", VISIBLE=False),
    h3_string_id("display_string"),
    h3_string_id("display_medal"),
    SEnum16("required_field", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience", *mulg_runtime_general_event_required_field),
    SEnum16("required_field_2", *mulg_runtime_general_event_required_field),
    SEnum16("excluded_audience_2", *mulg_runtime_general_event_required_field),
    h3_string_id("primary_string"),
    SInt32("primary_string_duration"),
    h3_string_id("plural_display_string"),
    Float("sound_delay_announcer_only"),
    Bool16("sound_flags",
        "announcer_sound",
        ),
    SInt16("unknown_3", VISIBLE=False),
    h3_dependency("english_sound"),
    h3_dependency("japanese_sound"),
    h3_dependency("german_sound"),
    h3_dependency("french_sound"),
    h3_dependency("spanish_sound"),
    h3_dependency("latin_american_spanish_sound"),
    h3_dependency("italian_sound"),
    h3_dependency("korean_sound"),
    h3_dependency("chinese_traditional_sound"),
    h3_dependency("chinese_simplified_sound"),
    h3_dependency("portuguese_sound"),
    h3_dependency("polish_sound"),
    BytesRaw("unknown_4", SIZE=16, VISIBLE=False),
    ENDIAN=">", SIZE=260
    )


mulg_runtime_multiplayer_constant_weapon = Struct("weapon",
    h3_dependency("weapon", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


mulg_runtime_multiplayer_constant_vehicle = Struct("vehicle",
    h3_dependency("vehicle", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=32
    )


mulg_runtime_multiplayer_constant_projectile = Struct("projectile",
    h3_dependency("projectile", VISIBLE=False),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=28
    )


mulg_runtime_multiplayer_constant_equipment = Struct("equipment",
    h3_dependency("equipment", VISIBLE=False),
    Float("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


mulg_runtime_multiplayer_constant = Struct("multiplayer_constant",
    Array("unknown_array_0", SUB_STRUCT=Float("unknown"), SIZE=26, VISIBLE=False),
    h3_reflexive("weapons", mulg_runtime_multiplayer_constant_weapon),
    h3_reflexive("vehicles", mulg_runtime_multiplayer_constant_vehicle),
    h3_reflexive("projectiles", mulg_runtime_multiplayer_constant_projectile),
    h3_reflexive("equipment", mulg_runtime_multiplayer_constant_equipment),
    Array("unknown_array_1", SUB_STRUCT=Float("unknown"), SIZE=40, VISIBLE=False),
    Float("maximum_random_spawn_bias"),
    Float("teleporter_recharge_time"),
    Float("grenade_danger_weight"),
    Float("grenade_danger_inner_radius"),
    Float("grenade_danger_outer_radius"),
    Float("grenade_danger_lead_time"),
    Float("vehicle_danger_minimum_speed"),
    Float("vehicle_danger_weight"),
    Float("vehicle_danger_radius"),
    Float("vehicle_danger_lead_time"),
    Float("vehicle_nearby_player_distance"),
    h3_dependency("hill_shader"),
    Array("unknown_array_2", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    h3_dependency("bomb_explode_effect"),
    h3_dependency("unknown_0"),
    h3_dependency("bomb_explode_damage_effect"),
    h3_dependency("bomb_defuse_effect"),
    h3_dependency("cursor_impact_effect"),
    h3_string_id("bomb_defusal_string"),
    h3_string_id("blocked_teleporter_string"),
    h3_string_id("unknown_1", VISIBLE=False),
    h3_string_id("spawn_allowed_default_respawn_string"),
    h3_string_id("spawn_at_player_looking_at_self_string"),
    h3_string_id("spawn_at_player_looking_at_target_string"),
    h3_string_id("spawn_at_player_looking_at_potential_target_string"),
    h3_string_id("spawn_at_territory_allowed_looking_at_target_string"),
    h3_string_id("spawn_at_territory_allowed_looking_at_potential_target_string"),
    h3_string_id("player_out_of_lives_string"),
    h3_string_id("invalid_spawn_target_string"),
    h3_string_id("targetted_player_enemies_nearby_string"),
    h3_string_id("targetted_player_unfriendly_team_string"),
    h3_string_id("targetted_player_is_dead_string"),
    h3_string_id("targetted_player_in_combat_string"),
    h3_string_id("targetted_player_too_far_from_owned_flag_string"),
    h3_string_id("no_available_netpoints_string"),
    h3_string_id("netpoint_contested_string"),
    ENDIAN=">", SIZE=540
    )


mulg_runtime_state_response = Struct("state_response",
    Bool16("flags", *unknown_flags_16),
    SInt16("unknown_0", VISIBLE=False),
    SEnum16("state", *mulg_runtime_state_response_state),
    SInt16("unknown_1", VISIBLE=False),
    h3_string_id("free_for_all_message"),
    h3_string_id("team_message"),
    h3_dependency("unknown_2"),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    ENDIAN=">", SIZE=36
    )


mulg_runtime = Struct("runtime",
    h3_dependency("sandbox_editor_unit"),
    h3_dependency("sandbox_editor_object"),
    h3_dependency("flag"),
    h3_dependency("ball"),
    h3_dependency("bomb"),
    h3_dependency("vip_zone"),
    h3_dependency("in_game_strings"),
    h3_reflexive("sounds", mulg_runtime_sound),
    h3_reflexive("looping_sounds", mulg_runtime_looping_sound),
    h3_reflexive("general_events", mulg_runtime_general_event),
    h3_reflexive("flavor_events", mulg_runtime_flavor_event),
    h3_reflexive("slayer_events", mulg_runtime_slayer_event),
    h3_reflexive("ctf_events", mulg_runtime_ctf_event),
    h3_reflexive("oddball_events", mulg_runtime_oddball_event),
    h3_reflexive("king_of_the_hill_events", mulg_runtime_king_of_the_hill_event),
    h3_reflexive("vip_events", mulg_runtime_vip_event),
    h3_reflexive("juggernaut_events", mulg_runtime_juggernaut_event),
    h3_reflexive("territories_events", mulg_runtime_territories_event),
    h3_reflexive("assault_events", mulg_runtime_assault_event),
    h3_reflexive("infection_events", mulg_runtime_infection_event),
    SInt32("default_frag_grenade_count"),
    SInt32("default_plasma_grenade_count"),
    h3_reflexive("multiplayer_constants", mulg_runtime_multiplayer_constant),
    h3_reflexive("state_responses", mulg_runtime_state_response),
    h3_dependency("scoreboard_emblem_bitmap"),
    h3_dependency("scoreboard_dead_emblem_bitmap"),
    h3_dependency("default_shape_shader"),
    h3_dependency("unknown"),
    h3_dependency("ctf_intro_ui"),
    h3_dependency("slayer_intro_ui"),
    h3_dependency("oddball_intro_ui"),
    h3_dependency("king_of_the_hill_intro_ui"),
    h3_dependency("sandbox_intro_ui"),
    h3_dependency("vip_intro_ui"),
    h3_dependency("juggernaut_intro_ui"),
    h3_dependency("territories_intro_ui"),
    h3_dependency("assault_intro_ui"),
    h3_dependency("infection_intro_ui"),
    ENDIAN=">", SIZE=524
    )


mulg_body = Struct("tagdata",
    h3_reflexive("universal", mulg_universal),
    h3_reflexive("runtime", mulg_runtime),
    ENDIAN=">", SIZE=24
    )


def get():
    return mulg_def

mulg_def = TagDef("mulg",
    h3_blam_header('mulg'),
    mulg_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["mulg"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'GeometryMesh', 'PermutationMesh', )


class GeometryMesh:
    verts = ()
    tris  = ()
    local_nodes = []
    def __init__(self, verts=(), tris=()):
        self.verts = verts if verts else []
        self.tris  = tris  if tris  else []


class PermutationMesh:
    markers = ()
    lod_meshes = ()
    is_random_perm = True

    def __init__(self):
        self.markers = []
        self.lod_meshes = {}



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from copy import copy, deepcopy
from math import pi
try:
    from mozzarilla.widgets.field_widgets import ReflexiveFrame,\
         HaloRawdataFrame, HaloUInt32ColorPickerFrame, TextFrame,\
         ColorPickerFrame, EntryFrame, HaloScriptSourceFrame,\
         SoundSampleFrame, DynamicArrayFrame, DynamicEnumFrame,\
         HaloScriptTextFrame, HaloBitmapTagFrame, FontCharacterFrame,\
         MeterImageFrame, HaloHudMessageTextFrame
except Exception:
    ReflexiveFrame = HaloRawdataFrame = HaloUInt32ColorPickerFrame =\
                     TextFrame = ColorPickerFrame = EntryFrame =\
                     HaloScriptSourceFrame = SoundSampleFrame =\
                     DynamicArrayFrame = DynamicEnumFrame =\
                     HaloScriptTextFrame = HaloBitmapTagFrame =\
                     FontCharacterFrame = MeterImageFrame =\
                     HaloHudMessageTextFrame = None

from supyr_struct.defs.common_descs import *
from supyr_struct.defs.block_def import BlockDef

from reclaimer.field_types import *
from reclaimer.field_type_methods import tag_ref_str_size,\
     read_string_id_string, write_string_id_string, get_set_string_id_size
from reclaimer.constants import *
from reclaimer.enums import *

# before we do anything, we need to inject these constants so any definitions
# that are built that use them will have them in their descriptor entries.
inject_halo_constants()

def tag_class(*args, **kwargs):
    '''
    A macro for creating a tag_class enum desc with the
    enumerations set to the provided tag_class fcc's.
    '''
    class_mapping = kwargs.pop("class_mapping", tag_class_fcc_to_ext)
    classes = tuple((class_mapping[fcc], fcc) for fcc in args)
    default = classes[0][1] if len(classes) == 1 else 0xffffffff

    return UEnum32(
        'tag_class',
        # NOTE: NONE is the expected enum_name in several places when checking
        # the value of an enumerator. DO NOT EVER change it or its letter case
        *(tuple(sorted(classes)) + (("NONE", 0xffffffff),) ),
        DEFAULT=default, GUI_NAME='', WIDGET_WIDTH=20, **kwargs
        )


def reflexive(name, substruct, max_count=MAX_REFLEXIVE_COUNT, *names, **desc):
    '''This function serves to macro the creation of a reflexive'''
    desc.update(
        INCLUDE=reflexive_struct,
        STEPTREE=Array(name + "_array",
            SIZE=".size", MAX=max_count,
            SUB_STRUCT=substruct, WIDGET=ReflexiveFrame
            ),
        SIZE=12
        )

    if DYN_NAME_PATH in desc:
        desc[STEPTREE][DYN_NAME_PATH] = desc.pop(DYN_NAME_PATH)
    if names:
        name_map = {}
        for i in range(len(names)):
            e_name = BlockDef.str_to_name(None, names[i])
            name_map[e_name] = i

        desc[STEPTREE][NAME_MAP] = name_map

    return Reflexive(name, **desc)

def get_raw_reflexive_offsets(desc, two_byte_offs, four_byte_offs, off=0):
    if INCLUDE in desc:
        desc = BlockDef.include_attrs(None, dict(desc))
        BlockDef.set_entry_count(None, desc)

    for i in range(desc.get(ENTRIES, 0)):
        f_desc = desc[i]
        f_type = f_desc[TYPE]
        size   = f_desc.get(SIZE, f_type.size)
        if f_type.is_data:
            big = f_type.big
            little = f_type.little
            if big is little or big.enc == little.enc:
                # endianness is forced. dont swap anything
                pass
            elif size == 4:
                four_byte_offs.append(off)
            elif size == 2:
                two_byte_offs.append(off)
            off += size
        elif f_type is Pad:
            off += size
        else:
            off = get_raw_reflexive_offsets(
                f_desc, two_byte_offs, four_byte_offs, off)
    return off


def raw_reflexive(name, substruct, max_count=MAX_REFLEXIVE_COUNT, *names, **desc):
    '''
    This function serves to macro the creation
    of a reflexive treated as rawdata
    '''
    # TODO: Make max_count be incorporated into desc and actually do something
    desc = reflexive(name, substruct, max_count, *names, **desc)
    info = desc[RAW_REFLEXIVE_INFO] = [0, [], []]
    sub_desc = desc[STEPTREE][SUB_STRUCT]
    sub_size = info[0] = sub_desc.get(SIZE)
    two_byte_offs, four_byte_offs = info[1], info[2]
    if sub_size is None:
        raise KeyError("Cannot make a raw_reflexive without a substruct size.")

    def data_size(parent=None, new_value=None, struct_size=sub_size, *a, **kw):
        if new_value is None:
            return parent.size*struct_size
        else:
            parent.size = new_value//struct_size

    desc[STEPTREE] = BytearrayRaw(desc[STEPTREE][NAME], SIZE=data_size)
    desc[STEPTREE][SUB_STRUCT] = sub_desc  # store this here in case its needed
    desc[TYPE] = RawReflexive
    get_raw_reflexive_offsets(sub_desc, two_byte_offs, four_byte_offs)
    return desc


def rawdata_ref(name, f_type=BytearrayRaw, max_size=None,
                widget=HaloRawdataFrame, **kwargs):
    '''This function serves to macro the creation of a rawdata reference'''
    ref_struct = dict(rawdata_ref_struct)
    if COMMENT in kwargs: ref_struct[COMMENT] = kwargs.pop(COMMENT)
    if TOOLTIP in kwargs: ref_struct[TOOLTIP] = kwargs.pop(TOOLTIP)
    if max_size is not None:
        ref_struct[0] = dict(ref_struct[0])
        ref_struct[0][MAX] = kwargs[MAX] = max_size

    if widget is not None:
        kwargs[WIDGET] = widget

    return RawdataRef(name,
        INCLUDE=ref_struct,
        STEPTREE=f_type("data", GUI_NAME="", SIZE=".size", **kwargs))


def rawtext_ref(name, f_type=StrRawLatin1, max_size=None,
                widget=TextFrame, **kwargs):
    '''This function serves to macro the creation of a rawdata reference'''
    ref_struct = dict(rawdata_ref_struct)
    kwargs.update(WIDGET=widget)
    ref_struct[0] = dict(ref_struct[0])
    ref_struct[0][VISIBLE] = VISIBILITY_METADATA
    if COMMENT in kwargs: ref_struct[COMMENT] = kwargs.pop(COMMENT)
    if TOOLTIP in kwargs: ref_struct[TOOLTIP] = kwargs.pop(TOOLTIP)
    if max_size is not None:
        ref_struct[0][MAX] = kwargs[MAX] = max_size

    return RawdataRef(name,
        INCLUDE=ref_struct, ORIENT="v",
        STEPTREE=f_type("data",
            SIZE=".size", GUI_NAME=name.replace('_', ' '), **kwargs)
            )


def get_meta_dependency_filepath(parent=None, tag_index_manager=None, **kwargs):
    if tag_index_manager and parent:
        tag_index_ref = tag_index_manager.get_tag_index_ref(parent.id)
        if tag_index_ref:
            parent.filepath = tag_index_ref.path
            return
    parent.filepath = ""


def dependency_uint32(name='tag_ref', **kwargs):
    kwargs.setdefault(VISIBLE, VISIBILITY_METADATA)
    kwargs.setdefault(ORIENT, "h")
    return QStruct(name,
        UInt32("id"),
        STEPTREE=Computed("filepath",
            COMPUTE_READ=get_meta_dependency_filepath, WIDGET_WIDTH=32
            ),
        **kwargs
        )


def dependency(name='tag_ref', valid_ids=None, **kwargs):
    '''This function serves to macro the creation of a tag dependency'''
    if isinstance(valid_ids, tuple):
        valid_ids = tag_class(*valid_ids)
    elif isinstance(valid_ids, str):
        valid_ids = tag_class(valid_ids)
    elif valid_ids is None:
        valid_ids = valid_tags

    return TagRef(name,
        valid_ids,
        INCLUDE=tag_ref_struct,
        STEPTREE=StrTagRef(
            "filepath", SIZE=tag_ref_str_size, GUI_NAME="", MAX=254),
        **kwargs
        )


def zone_asset(name, **kwargs):
    return ZoneAsset(name, INCLUDE=zone_asset_struct, **kwargs)


def string_id(name, index_bit_ct, set_bit_ct, len_bit_ct=None, **kwargs):
    if len_bit_ct is None:
        len_bit_ct = 32 - index_bit_ct - set_bit_ct

    kwargs.setdefault(STRINGID_IDX_BITS, index_bit_ct)
    kwargs.setdefault(STRINGID_SET_BITS, set_bit_ct)
    kwargs.setdefault(STRINGID_LEN_BITS, len_bit_ct)
    kwargs.setdefault(ORIENT, "h")
    return StringID(name,
        UInt32('string_id', VISIBLE=VISIBILITY_METADATA),
        STEPTREE=WritableComputed("string",
            COMPUTE_READ=read_string_id_string,
            COMPUTE_WRITE=write_string_id_string,
            COMPUTE_SIZECALC=(lambda node, **kw: len(node) + bool(node)),
            SIZE=get_set_string_id_size, GUI_NAME="",
            WIDGET_WIDTH=32, NODE_CLS=str, MAX=(1 << len_bit_ct)
            ),
        **kwargs
        )


def blam_header(tagid, version=1):
    '''This function serves to macro the creation of a tag header'''
    header_desc = dict(tag_header)
    header_desc[1] = dict(header_desc[1])
    header_desc[5] = dict(header_desc[5])
    header_desc[1][DEFAULT] = tagid
    header_desc[5][DEFAULT] = version
    return header_desc


_func_unit_scales = {}  # holds created unit_scales for reuse

def get_unit_scale(scale60, val30=1):
    """
    Returns a function to be used in the UNIT_SCALE of a descriptor that
    will change the scale depending on if the 60fps flag is set or not.
    """
    assert 0 not in (val30, scale60), ("60fps scale and default 30fps " +
                                       "value must both be non-zero.")
    val60 = val30*scale60

    key = (val30, val60)
    if key in _func_unit_scales:
        return _func_unit_scales[key]

    def unit_scale(*args, **kwargs):
        w = kwargs.get('f_widget')
        if kwargs.get('get_scales'):
            # used for getting both the 30 and 60 fps scales
            return (val30, val60)

        try:
            if w.tag_window.save_as_60:
                return val60
        except AttributeError:
            return val30

    _func_unit_scales[key] = unit_scale
    unit_scale.fps_60_scale = True
    return unit_scale

# typical unit_scales for seconds and seconds squared fields
per_sec_unit_scale    = get_unit_scale(2)
per_sec_sq_unit_scale = get_unit_scale(4)
sec_unit_scale    = get_unit_scale(1/2)
sec_sq_unit_scale = get_unit_scale(1/4)

irad = 180/pi
irad_per_sec_unit_scale = get_unit_scale(1/2, irad)
irad_per_sec_sq_unit_scale = get_unit_scale(1/4, irad)

def dyn_senum8(name, *args, **kwargs):
    kwargs.setdefault('DEFAULT', -1)
    kwargs.setdefault('WIDGET', DynamicEnumFrame)
    return SInt8(name, *args, **kwargs)

def dyn_senum16(name, *args, **kwargs):
    kwargs.setdefault('DEFAULT', -1)
    kwargs.setdefault('WIDGET', DynamicEnumFrame)
    return SInt16(name, *args, **kwargs)

def dyn_senum32(name, *args, **kwargs):
    kwargs.setdefault('DEFAULT', -1)
    kwargs.setdefault('WIDGET', DynamicEnumFrame)
    return SInt32(name, *args, **kwargs)

def ascii_str32(name, **kwargs):
    # encoding used is latin1 to take care of cases
    # where the string has invalid characters in it
    return StrLatin1(str(name), SIZE=32, **kwargs)

def float_zero_to_one(name, *args, **kwargs):
    return Float(name, *args, MIN=0.0, MAX=1.0, SIDETIP="[0,1]", **kwargs)
def float_neg_one_to_one(name, *args, **kwargs):
    return Float(name, *args, MIN=-1.0, MAX=1.0, SIDETIP="[-1,1]", **kwargs)

def float_sec(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "seconds")
    kwargs.setdefault('UNIT_SCALE', sec_unit_scale)
    return Float(name, *args, **kwargs)

def float_deg(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "degrees")
    return Float(name, *args, **kwargs)

def float_deg_sec(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "degrees/sec")
    kwargs.setdefault('UNIT_SCALE', per_sec_unit_scale)
    return Float(name, *args, **kwargs)

def float_rad(name, *args, **kwargs):
    return Float(name, *args, SIDETIP="degrees",
                 UNIT_SCALE=irad, **kwargs)
def float_rad_sec(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "degrees/sec")
    kwargs.setdefault('UNIT_SCALE', irad_per_sec_unit_scale)
    return Float(name, *args, **kwargs)
def float_rad_sec_sq(name, *args, **kwargs):
    # yes, keep this as per_sec and not per_sec_sq.
    # this is because of how halo uses these values.
    kwargs.setdefault('SIDETIP', "degrees/(sec^2)")
    kwargs.setdefault('UNIT_SCALE', irad_per_sec_unit_scale)
    return Float(name, *args, **kwargs)

def float_wu(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "world units")
    return Float(name, *args, **kwargs)
def float_wu_sec(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "world units/sec")
    kwargs.setdefault('UNIT_SCALE', per_sec_unit_scale)
    return Float(name, *args, **kwargs)
def float_wu_sec_sq(name, *args, **kwargs):
    kwargs.setdefault('SIDETIP', "world units/(sec^2)")
    kwargs.setdefault('UNIT_SCALE', per_sec_sq_unit_scale)
    return Float(name, *args, **kwargs)

def float_zero_to_inf(name, *args, **kwargs):
    return Float(name, *args, SIDETIP="[0,+inf]", **kwargs)


from_to_float = QStruct('',
    Float("from", GUI_NAME=''), Float("to"), ORIENT='h'
    )
from_to = from_to_float
from_to_sint32 = QStruct('',
    SInt32("from", GUI_NAME=''), SInt32("to"), ORIENT='h'
    )
from_to_sint16 = QStruct('',
    SInt16("from", GUI_NAME=''), SInt16("to"), ORIENT='h'
    )
from_to_sint8 = QStruct('',
    SInt8("from", GUI_NAME=''), SInt8("to"), ORIENT='h'
    )
from_to_uint32 = QStruct('',
    UInt32("from", GUI_NAME=''), UInt32("to"), ORIENT='h'
    )
from_to_uint16 = QStruct('',
    UInt16("from", GUI_NAME=''), UInt16("to"), ORIENT='h'
    )
from_to_uint8 = QStruct('',
    UInt8("from", GUI_NAME=''), UInt8("to"), ORIENT='h'
    )

def color_argb_uint32(name, **kwargs):
    return QStruct(name, INCLUDE=argb_byte, **kwargs)

def color_xrgb_uint32(name, **kwargs):
    return QStruct(name, INCLUDE=xrgb_byte, **kwargs)

def color_argb_float(name, **kwargs):
    return QStruct(name, INCLUDE=argb_float, **kwargs)

def color_rgb_float(name, **kwargs):
    return QStruct(name, INCLUDE=rgb_float, **kwargs)

def from_to_deg(name, *args, **kwargs):
    return QStruct(name,
        Float("from", GUI_NAME=''),
        Float("to"), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def from_to_rad(name, *args, **kwargs):
    return QStruct(name,
        Float("from", UNIT_SCALE=irad, GUI_NAME=''),
        Float("to",   UNIT_SCALE=irad), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def from_to_rad_sec(name, *args, **kwargs):
    return QStruct(name,
        Float("from", UNIT_SCALE=irad_per_sec_unit_scale, GUI_NAME=''),
        Float("to",   UNIT_SCALE=irad_per_sec_unit_scale), *args,
        ORIENT='h', SIDETIP='degrees/sec', **kwargs
        )

def from_to_sec(name, *args, **kwargs):
    return QStruct(name,
        Float("from", UNIT_SCALE=sec_unit_scale, GUI_NAME=''),
        Float("to",   UNIT_SCALE=sec_unit_scale), *args,
        ORIENT='h', SIDETIP='seconds', **kwargs
        )

def from_to_wu(name, *args, **kwargs):
    return QStruct(name,
        Float("from", GUI_NAME=''),
        Float("to"), *args,
        ORIENT='h', SIDETIP='world units', **kwargs
        )

def from_to_wu_sec(name, *args, **kwargs):
    return QStruct(name,
        Float("from", UNIT_SCALE=per_sec_unit_scale, GUI_NAME=''),
        Float("to",   UNIT_SCALE=per_sec_unit_scale), *args,
        ORIENT='h', SIDETIP='world units/sec', **kwargs
        )

def from_to_zero_to_one(name, *args, **kwargs):
    return QStruct(name,
        Float("from", MIN=0.0, MAX=1.0, GUI_NAME=''),
        Float("to", MIN=0.0, MAX=1.0), *args,
        ORIENT='h', SIDETIP='[0,1]', **kwargs
        )

def from_to_neg_one_to_one(name, *args, **kwargs):
    return QStruct(name,
        Float("from", MIN=-1.0, MAX=1.0, GUI_NAME=''),
        Float("to", MIN=-1.0, MAX=1.0), *args,
        ORIENT='h', SIDETIP='[-1,1]', **kwargs
        )

def yp_float_deg(name, *args, **kwargs):
    return QStruct(name,
        Float("y"), Float("p"), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def ypr_float_deg(name, *args, **kwargs):
    return QStruct(name,
        Float("y"), Float("p"), Float("r"), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def yp_float_rad(name, *args, **kwargs):
    return QStruct(name,
        Float("y", UNIT_SCALE=irad),
        Float("p", UNIT_SCALE=irad), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def ypr_float_rad(name, *args, **kwargs):
    return QStruct(name,
        Float("y", UNIT_SCALE=irad),
        Float("p", UNIT_SCALE=irad),
        Float("r", UNIT_SCALE=irad), *args,
        ORIENT='h', SIDETIP='degrees', **kwargs
        )

def anim_func_per_pha_macro(name, **desc):
    desc.setdefault("HIDE_TITLE", True)
    return Struct(name,
        SEnum16("function", *animation_functions,
            GUI_NAME="%s function" % name),
        Pad(2),
        Float("period", SIDETIP='seconds',
            GUI_NAME="%s period" % name, UNIT_SCALE=sec_unit_scale),
        Float("phase",  SIDETIP='seconds',
            GUI_NAME="%s phase" % name,  UNIT_SCALE=sec_unit_scale),
        **desc
        )

def anim_func_per_sca_macro(name, **desc):
    desc.setdefault("HIDE_TITLE", True)
    return Struct(name,
        SEnum16("function", *animation_functions,
            GUI_NAME="%s function" % name),
        Pad(2),
        Float("period", SIDETIP='seconds',
            GUI_NAME="%s period" % name, UNIT_SCALE=sec_unit_scale),
        Float("scale",  SIDETIP='base map repeats',
            GUI_NAME="%s scale" % name),  # base map repeats
        **desc
        )

def anim_src_func_per_pha_sca_macro(name, **desc):
    desc.setdefault("HIDE_TITLE", True)
    return Struct(name,
        SEnum16("source", *function_outputs,
            GUI_NAME="%s source" % name),
        SEnum16("function", *animation_functions,
            GUI_NAME="%s function" % name),
        Float("period", SIDETIP='seconds',
            GUI_NAME="%s period" % name, UNIT_SCALE=sec_unit_scale),
        Float("phase", SIDETIP='seconds',
            GUI_NAME="%s phase" % name, UNIT_SCALE=sec_unit_scale),
        Float("scale",  SIDETIP='repeats',
            GUI_NAME="%s scale" % name),  # repeats
        **desc
        )

def anim_src_func_per_pha_sca_rot_macro(name, **desc):
    desc.setdefault("HIDE_TITLE", True)
    return Struct(name,
        SEnum16("source", *function_outputs,
            GUI_NAME="%s source" % name),
        SEnum16("function", *animation_functions,
            GUI_NAME="%s function" % name),
        Float("period", SIDETIP='seconds',
            GUI_NAME="%s period" % name, UNIT_SCALE=sec_unit_scale),
        Float("phase", SIDETIP='seconds',
            GUI_NAME="%s phase" % name, UNIT_SCALE=sec_unit_scale),
        Float("scale",  SIDETIP='degrees',
            GUI_NAME="%s scale" % name),  # repeats
        **desc
        )


valid_tags = tag_class(*sorted(tag_class_fcc_to_ext.keys()))
valid_models = tag_class('mode', 'mod2')
valid_event_effects = tag_class('effe', 'snd!')
valid_attachments = tag_class('cont', 'effe', 'ligh', 'mgs2', 'pctl', 'lsnd')
valid_effect_events = tag_class(
    'bipd', 'jpt!', 'deca', 'devi', 'ctrl', 'lifi', 'mach',
    'eqip', 'garb', 'item', 'ligh', 'obje', 'pctl', 'plac',
    'proj', 'scen', 'snd!', 'ssce', 'unit', 'vehi', 'weap'
    )
valid_items = tag_class('eqip', 'garb', 'item', 'weap')
valid_objects = tag_class(
    'bipd', 'ctrl', 'lifi', 'mach', 'eqip', 'garb', 'obje',
    'plac', 'proj', 'scen', 'ssce', 'vehi', 'weap'
    )
valid_units = tag_class('bipd', 'unit', 'vehi')
valid_devices_items_objects_units = tag_class(
    'bipd', 'devi', 'ctrl', 'lifi', 'mach', 'eqip', 'garb', 'item',
    'obje', 'plac', 'proj', 'scen', 'ssce', 'unit', 'vehi', 'weap'
    )
valid_shaders = tag_class(
    'shdr', 'schi', 'scex', 'sotr', 'senv',
    'sgla', 'smet', 'soso', 'spla', 'swat'
    )
valid_widgets = tag_class('ant!', 'flag', 'glw!', 'mgs2', 'elec')



# Descriptors
tag_header = Struct("blam_header",
    Pad(36),
    UEnum32("tag_class",
        GUI_NAME="tag_class", INCLUDE=valid_tags, EDITABLE=False
        ),
    UInt32("checksum", DEFAULT=0, EDITABLE=False),
    UInt32("header_size",  DEFAULT=64, EDITABLE=False),
    BBool64("flags",
        "edited_with_mozz",
        EDITABLE=False
        ),
    UInt16("version", DEFAULT=1, EDITABLE=False),
    UInt8("integrity0", DEFAULT=0, EDITABLE=False),
    UInt8("integrity1", DEFAULT=255, EDITABLE=False),
    UEnum32("engine_id",
        ("halo_1", 'blam'),
        DEFAULT='blam', EDITABLE=False
        ),
    VISIBLE=VISIBILITY_METADATA, SIZE=64, ENDIAN=">"  # KEEP THE ENDIAN SPECIFICATION
    )

# Miscellaneous, Halo specific descriptors
anim_func_per_pha = Struct('',
    SEnum16("function", *animation_functions),
    Pad(2),
    Float("period", SIDETIP='seconds', UNIT_SCALE=sec_unit_scale),
    Float("phase", SIDETIP='seconds',  UNIT_SCALE=sec_unit_scale),
    )

anim_func_per_sca = Struct('',
    SEnum16("function", *animation_functions),
    Pad(2),
    Float("period", SIDETIP='seconds', UNIT_SCALE=sec_unit_scale),
    Float("scale", SIDETIP='base map repeats'),  # base map repeats
    )

anim_src_func_per_pha_sca = Struct('',
    SEnum16("source", *function_outputs),
    SEnum16("function", *animation_functions),
    Float("period", SIDETIP='seconds', UNIT_SCALE=sec_unit_scale),
    Float("phase", SIDETIP='seconds',  UNIT_SCALE=sec_unit_scale),
    Float("scale", SIDETIP='repeats'),  # repeats
    )

anim_src_func_per_pha_sca_rot = Struct('',
    SEnum16("source", *function_outputs),
    SEnum16("function", *animation_functions),
    Float("period", SIDETIP='seconds', UNIT_SCALE=sec_unit_scale),
    Float("phase", SIDETIP='seconds',  UNIT_SCALE=sec_unit_scale),
    Float("scale", SIDETIP='degrees'),  # repeats
    )


# This is the descriptor used wherever a tag references a rawdata chunk
rawdata_ref_struct = RawdataRef('rawdata_ref',
    SInt32("size", GUI_NAME="", SIDETIP="bytes", EDITABLE=False),
    Bool32("flags",
        "data_in_resource_map",
        VISIBLE=VISIBILITY_METADATA, EDITABLE=False
        ),
    UInt32("raw_pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),  # doesnt use magic
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    ORIENT='h'
    )

# This is the descriptor used wherever a tag reference a reflexive
reflexive_struct = Reflexive('reflexive',
    SInt32("size", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),  # 0 in meta it seems
    )

# This is the descriptor used wherever a tag references another tag
tag_ref_struct = TagRef('dependency',
    valid_tags,
    SInt32("path_pointer", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    SInt32("path_length", MAX=MAX_TAG_PATH_LEN, VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    UInt32("id", VISIBLE=VISIBILITY_METADATA, EDITABLE=False),
    ORIENT='h'
    )

predicted_resource = Struct('predicted_resource',
    SInt16('type',
        'bitmap',
        'sound',
        ),
    SInt16('resource_index'),
    UInt32('tag_index'),
    )

zone_asset_struct = ZoneAsset("zone_asset",
    UInt16("salt"),
    UInt16("idx"),
    UInt32("unused", VISIBLE=VISIBILITY_METADATA),
    )

extra_layers_block = dependency("extra_layer", valid_shaders)

damage_modifiers = QStruct("damage_modifiers",
    *(float_zero_to_inf(material_name) for material_name in materials_list)
    )

# Miscellaneous shared descriptors
compressed_normal_32 = BitStruct('compressed_norm32',
    S1BitInt("i", SIZE=11, UNIT_SCALE=1/1023, MIN=-1023, MAX=1023, WIDGET_WIDTH=10),
    S1BitInt("j", SIZE=11, UNIT_SCALE=1/1023, MIN=-1023, MAX=1023, WIDGET_WIDTH=10),
    S1BitInt("k", SIZE=10, UNIT_SCALE=1/511,  MIN=-511,  MAX=511,  WIDGET_WIDTH=10),
    ORIENT='h'
    )

# coordinates
xyz_float = QStruct('xyz_float',
    Float("x"), Float("y"), Float("z"), ORIENT='h'
    )
xy_float = QStruct('xy_float',
    Float("x"), Float("y"), ORIENT='h'
    )

# colors
argb_float = QStruct('argb_float',
    Float("a", MIN=0.0, MAX=1.0),
    Float("r", MIN=0.0, MAX=1.0),
    Float("g", MIN=0.0, MAX=1.0),
    Float("b", MIN=0.0, MAX=1.0),
    ORIENT='h', WIDGET=ColorPickerFrame
    )
rgb_float = QStruct('rgb_float',
    Float("r", MIN=0.0, MAX=1.0),
    Float("g", MIN=0.0, MAX=1.0),
    Float("b", MIN=0.0, MAX=1.0),
    ORIENT='h', WIDGET=ColorPickerFrame
    )
xrgb_byte = QStruct('xrgb_uint8',
    Pad(1), UInt8("r"), UInt8("g"), UInt8("b"),
    ORIENT='h', WIDGET=ColorPickerFrame
    )

argb_byte = QStruct('argb_uint8',
    UInt8("a"), UInt8("r"), UInt8("g"), UInt8("b"),
    ORIENT='h', WIDGET=ColorPickerFrame
    )

xrgb_uint32 = UInt32('xrgb_uint32',
    WIDGET=HaloUInt32ColorPickerFrame, COLOR_CHANNELS="rgb", ORIENT="h")
argb_uint32 = UInt32('argb_uint32',
    WIDGET=HaloUInt32ColorPickerFrame, COLOR_CHANNELS="argb", ORIENT="h")

# rotations
ijkw_sint16 = QStruct('ijkw_sint16',
    SInt16("i", UNIT_SCALE=1/32767), SInt16("j", UNIT_SCALE=1/32767),
    SInt16("k", UNIT_SCALE=1/32767), SInt16("w", UNIT_SCALE=1/32767,
                                            DEFAULT=32767),
    ORIENT='h'
    )
ijk_sint16 = QStruct('ijk_sint16',
    SInt16("i", UNIT_SCALE=1/32767), SInt16("j", UNIT_SCALE=1/32767),
    SInt16("k", UNIT_SCALE=1/32767),
    ORIENT='h'
    )
ij_sint16 = QStruct('ij_sint16',
    SInt16("i", UNIT_SCALE=1/32767), SInt16("j", UNIT_SCALE=1/32767),
    ORIENT='h'
    )
ijkw_float = QStruct('ijkw_float',
    Float("i"), Float("j"), Float("k"), Float("w", DEFAULT=1.0),
    ORIENT='h'
    )
ijk_float = QStruct('ijk_float',
    Float("i"), Float("j"), Float("k"),
    ORIENT='h'
    )
ij_float = QStruct('ij_float',
    Float("i"), Float("j"),
    ORIENT='h'
    )
yp_float = QStruct('yp_float',
    Float("y"), Float("p"),
    ORIENT='h',
    )
ypr_float = QStruct('ypr_float',
    Float("y"), Float("p"), Float("r"),
    ORIENT='h',
    )

# texture coordinates
uv_float = QStruct('uv_float',
    Float("u"), Float("v"),
    ORIENT='h'
    )

#############################
# Open Sauce related things #
#############################
def tag_class_os(*args, **kwargs):
    '''
    A macro for creating an Open Sauce tag_class enum desc
    with the enumerations set to the provided tag_class fcc's.
    '''
    kwargs["class_mapping"] = tag_class_fcc_to_ext_os
    return tag_class(*args, **kwargs)


def dependency_os(name='tag_ref', valid_ids=None, **kwargs):
    '''This function serves to macro the creation of a tag dependency'''
    if isinstance(valid_ids, tuple):
        valid_ids = tag_class_os(*valid_ids)
    elif isinstance(valid_ids, str):
        valid_ids = tag_class_os(valid_ids)
    elif valid_ids is None:
        valid_ids = valid_tags_os

    return TagRef(name,
        valid_ids,
        INCLUDE=tag_ref_struct,
        STEPTREE=StrTagRef(
            "filepath", SIZE=tag_ref_str_size, GUI_NAME="", MAX=254),
        **kwargs
        )


def blam_header_os(tagid, version=1):
    '''This function serves to macro the creation of a tag header'''
    header_desc = dict(tag_header_os)
    header_desc[1] = dict(header_desc[1])
    header_desc[5] = dict(header_desc[5])
    header_desc[1][DEFAULT] = tagid
    header_desc[5][DEFAULT] = version
    return header_desc


valid_tags_os = tag_class_os(*sorted(tag_class_fcc_to_ext_os.keys()))

# i'm so cheeky, look at me abusing Courier New
blam_engine_id = UEnum32("engine_id",
    ("halo_1", 'blam'),
    ("halo_2", 'b2am'),
    ("halo_3", 'b3am'),
    ("halo_4", 'b4am'),
    ("halo_5", 'b5am'),
    ("halo_odst", 'ODST'),
    ("halo_reach", 'bRam'),
    ("halo_reach_beta", 'bRBm'),
    ("halo_4_net_test", 'b4Tm'),
    DEFAULT='blam', EDITABLE=False
    )

# Descriptors
tag_header_os = Struct("blam_header",
    Pad(36),
    UEnum32("tag_class",
        GUI_NAME="tag_class", INCLUDE=valid_tags_os, EDITABLE=False
        ),
    UInt32("checksum", DEFAULT=0, EDITABLE=False),
    UInt32("header_size", DEFAULT=64, EDITABLE=False),
    Bool64("flags",
        "edited_with_mozz",
        EDITABLE=False
        ),
    UInt16("version", DEFAULT=1, EDITABLE=False),
    UInt8("integrity0", DEFAULT=0, EDITABLE=False),
    UInt8("integrity1", DEFAULT=255, EDITABLE=False),
    blam_engine_id,
    VISIBLE=VISIBILITY_METADATA, SIZE=64, ENDIAN=">"  # KEEP THE ENDIAN SPECIFICATION
    )

valid_model_animations_yelo = tag_class_os('antr', 'magy')

os_shader_extension = Struct("os_shader_extension",
    Pad(4),
    QStruct("fade_controls",
        Float("depth_fade_distance"),
        Float("camera_fade_distance"),
        COMMENT="Controls the softness of an effect"
        ),
    SIZE=48
    )


def make_dependency_block(class_name="NONE", tag_id=0xFFffFFff,
                          tag_path="", tag_path_pointer=0xFFffFFff,
                          block_def=BlockDef(dependency())):
    block = block_def.build()
    try:
        block.tag_class.set_to(class_name)
    except Exception:
        pass
    block.id = tag_id & 0xFFffFFff
    block.path_pointer = tag_path_pointer & 0xFFffFFff
    block.filepath = str(tag_path)
    return block


def make_dependency_os_block(class_name="NONE", tag_id=0xFFffFFff,
                             tag_path="", tag_path_pointer=0xFFffFFff,
                             block_def=BlockDef(dependency_os())):
    return make_dependency_block(class_name, tag_id, tag_path,
                                 tag_path_pointer, block_def)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = ( 'MergedJmsRegion', )


from .mesh import ( PermutationMesh, GeometryMesh, )
from .model import JmsModel
from .triangle import JmsTriangle
from .vertex import JmsVertex


class MergedJmsRegion:
    name = ""
    perm_meshes = ()
    _split_by_shader = True

    def __init__(self, name, *jms_models, split_by_shader=True):
        self.name = name
        self._split_by_shader = split_by_shader
        self.perm_meshes = {}

        for jms_model in jms_models:
            self.merge_jms_model(jms_model)

    def merge_jms_model(self, jms_model, merged_jms_materials):
        assert isinstance(jms_model, JmsModel)
        try:
            src_region_index = jms_model.regions.index(self.name)
        except ValueError:
            # this region is not in the jms model provided
            return

        lod_level = jms_model.lod_level
        perm_name = jms_model.perm_name

        if perm_name not in self.perm_meshes:
            self.perm_meshes[perm_name] = PermutationMesh()
            self.perm_meshes[perm_name].is_random_perm = jms_model.is_random_perm

        perm_mesh = self.perm_meshes[perm_name]
        # copy the markers from the JmsModel we're given for this region,
        # BUT only do so if the lod_level is superhigh(this is what tool
        # does, and it makes sense to do it this way to prevent duplicates).
        if jms_model.lod_level == "superhigh":
            for marker in jms_model.markers:
                if marker.region == src_region_index:
                    perm_mesh.markers.append(marker)

        mesh_data = perm_mesh.lod_meshes.setdefault(lod_level, {})

        src_verts = jms_model.verts
        src_tris  = jms_model.tris
        region_verts = []
        region_tris = [None] * len(src_tris)

        mat_indices_by_name, i = {}, 0
        for mat in merged_jms_materials:
            mat_indices_by_name.setdefault(mat.name, []).append(i)
            i += 1

        mat_map, i = [0] * len(jms_model.materials), 0
        for mat in jms_model.materials:
            mat_map[i] = mat_indices_by_name[mat.name].pop(0)
            i += 1

        vert_map = dict()
        get_add_vert = vert_map.setdefault
        v_base = len(region_verts)
        tri_ct = 0
        mat_nums = set()
        for tri in src_tris:
            if tri.region == src_region_index:
                mat_num = mat_map[tri.shader]
                # region number doesnt matter at this point for triangles
                # since it isnt stored in compiled models, so set it to -1
                tri = JmsTriangle(-1, mat_num, tri.v0, tri.v1, tri.v2)
                tri.v0 = get_add_vert(tri.v0, v_base + len(vert_map))
                tri.v1 = get_add_vert(tri.v1, v_base + len(vert_map))
                tri.v2 = get_add_vert(tri.v2, v_base + len(vert_map))
                mat_nums.add(mat_num)
                region_tris[tri_ct] = tri
                tri_ct += 1

        if tri_ct == 0:
            return

        # collect all the verts and triangles used by this region
        region_verts.extend([None] * len(vert_map))
        for i, j in vert_map.items():
            v = src_verts[i]
            region_verts[j] = JmsVertex(
                v.node_0, v.pos_x, v.pos_y, v.pos_z,
                v.norm_i, v.norm_j, v.norm_k, v.node_1, v.node_1_weight,
                v.tex_u, v.tex_v, v.tex_w,
                v.binorm_i,  v.binorm_j,  v. binorm_k,
                v.tangent_i, v.tangent_j, v.tangent_k)

        del region_tris[tri_ct: ]

        if len(mat_nums) == 1 or not self._split_by_shader:
            for mat_num in mat_nums: break
            if not self._split_by_shader:
                mat_num = -1

            mesh_data[mat_num] = GeometryMesh()
            mesh_data[mat_num].verts = region_verts
            mesh_data[mat_num].tris  = region_tris
            return

        # make a mesh for each material
        for mat_num in mat_nums:
            if mat_num not in mesh_data:
                mesh_data[mat_num] = GeometryMesh()

            mat_verts = mesh_data[mat_num].verts
            mat_tris  = mesh_data[mat_num].tris

            vert_map = dict()
            get_add_vert = vert_map.setdefault
            v_base = len(mat_verts)
            tri_ct = 0
            for tri in region_tris:
                if tri.shader == mat_num:
                    tri.v0 = get_add_vert(tri.v0, v_base + len(vert_map))
                    tri.v1 = get_add_vert(tri.v1, v_base + len(vert_map))
                    tri.v2 = get_add_vert(tri.v2, v_base + len(vert_map))
                    tri_ct += 1

            mat_verts.extend([None] * len(vert_map))
            for i, j in vert_map.items():
                mat_verts[j] = region_verts[i]

            i = len(mat_tris)
            mat_tris.extend([None] * tri_ct)
            for tri in region_tris:
                if tri.shader == mat_num:
                    mat_tris[i] = tri
                    i += 1



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.lens import LensTag
from supyr_struct.defs.tag_def import TagDef

occlusion_comment = """Occlusion factor affects overall lens flare brightness and can also affect scale.
Occlusion never affects rotation."""

corona_rotation_comment = """Controls how corona rotation is affected by the viewer and light angles."""

reflection = Struct("reflection",
    Bool16("flags",
        "align_rotation_with_screen_center",
        "radius_not_scaled_by_distance",
        "radius_scaled_by_occlusion_factor",
        "occluded_by_solid_objects",
        ),
    Pad(2),
    SInt16("bitmap_index"),
    Pad(22),
    Float("position", SIDETIP="along flare axis"), # along flare axis
    float_deg("rotation_offset"), # degrees
    Pad(4),
    from_to_wu("radius"),  # world units
    SEnum16("radius_scaled_by",
        "none",
        "rotation",
        "rotation_and_strafing",
        "distance_from_center",
        ),
    Pad(2),
    from_to_zero_to_one("brightness"),  # [0,1]
    SEnum16("brightness_scaled_by",
        "none",
        "rotation",
        "rotation_and_strafing",
        "distance_from_center",
        ),
    Pad(2),

    #Tint color
    QStruct("tint_color", INCLUDE=argb_float),

    #Animation
    Struct("animation",
        QStruct("color_lower_bound", INCLUDE=argb_float),
        QStruct("color_upper_bound", INCLUDE=argb_float),
        Bool16("flags", *blend_flags),
        SEnum16("function", *animation_functions),
        float_sec("period"),  # seconds
        float_sec("phase"),  # seconds
        ),

    SIZE=128
    )


lens_body = Struct("tagdata",
    float_rad("falloff_angle"),  # radians
    float_rad("cutoff_angle"),  # radians
    FlFloat("cosine_falloff_angle", VISIBLE=False),
    FlFloat("cosine_cutoff_angle", VISIBLE=False),
    Struct("occlusion",
        float_wu("radius"),
        SEnum16("offset_direction",
            "toward_viewer",
            "marker_forward",
            "none",
            ),
        Pad(2),
        float_wu("near_fade_distance"),
        float_wu("far_fade_distance"),
	    COMMENT=occlusion_comment
        ),

    Struct("bitmaps",
        dependency("bitmap", "bitm"),
        Bool16("flags",
            "sun",
            ),
        Pad(78),
        ),

    Struct("corona_rotation",
        SEnum16("function",
            "none",
            "rotation_a",
            "rotation_b",
            "rotation_translation",
            "translation",
            ),
        Pad(2),
        float_rad("function_scale"),  # radians
		COMMENT=corona_rotation_comment
        ),

    Struct("corona_radius_scale",
        Pad(24),
        Float("horizontal_scale"),
        Float("vertical_scale"),
        ),

    Pad(28),

    reflexive("reflections", reflection, 32),

    SIZE=240,
    )


def get():
    return lens_def

lens_def = TagDef("lens",
    blam_header("lens", 2),
    lens_body,

    ext=".lens_flare", endian=">", tag_cls=LensTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Copypasta H2.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

pmov_movement_type = (
    "physics",
    "collider",
    "swarm",
    "wind",
    )


pmov_movement_parameter = Struct("parameter",
    SInt32("parameter_id"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("unknown_1", VISIBLE=False),
    h3_rawdata_ref("function"),
    Float("unknown_2", VISIBLE=False),
    UInt8("unknown_3", VISIBLE=False),
    SInt8("unknown_4", VISIBLE=False),
    SInt8("unknown_5", VISIBLE=False),
    SInt8("unknown_6", VISIBLE=False),
    ENDIAN=">", SIZE=36
    )


pmov_movement = Struct("movement",
    SEnum16("type", *pmov_movement_type),
    SInt16("unknown_0", VISIBLE=False),
    h3_reflexive("parameters", pmov_movement_parameter),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    SInt32("unknown_3", VISIBLE=False),
    ENDIAN=">", SIZE=24
    )


pmov_body = Struct("tagdata",
    h3_dependency("template"),
    Bool32("flags",
        "physics",
        "collide_with_structure",
        "collide_with_media",
        "collide_with_scenery",
        "collide_with_vehicles",
        "collide_with_bipeds",
        "swarm",
        "wind",
        ),
    h3_reflexive("movements", pmov_movement),
    ENDIAN=">", SIZE=32
    )


def get():
    return pmov_def

pmov_def = TagDef("pmov",
    h3_blam_header('pmov'),
    pmov_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["pmov"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef


control_point = Struct("control_point",
    QStruct("position", INCLUDE=ijk_float),
    QStruct("orientation", INCLUDE=ijkw_float),
    SIZE=60,
    )

trak_body = Struct("tagdata",
    Pad(4),
    reflexive("control_points", control_point, 16),
    SIZE=48,
    )


def get():
    return trak_def

trak_def = TagDef("trak",
    blam_header('trak', 2),
    trak_body,

    ext=".camera_track", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from supyr_struct.defs.tag_def import BlockDef
from reclaimer.field_types import Container, FlUInt16, BytesRaw, WhileArray


def linesize(parent=None, new_value=None, *args, **kwargs):
    if parent is None:
        raise KeyError
    if new_value is None:
        return 4*parent.width

    parent.width = new_value//4


def has_next_line(rawdata=None, *args, **kwargs):
    try:
        return len(rawdata.peek(6)) >= 6
    except Exception:
        return False

meter_line = Container("meter_line",
    FlUInt16("x_pos"),
    FlUInt16("y_pos"),
    FlUInt16("width"),
    BytesRaw("line_data", SIZE=linesize)
    )

meter_image = WhileArray("meter_image",
    CASE=has_next_line,
    SUB_STRUCT=meter_line
    )

meter_image_def = BlockDef(meter_image)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .efpp import *
from ...hek.defs.objs.tag import HekTag

shader = Struct("shader",
    dependency_os("shader", 'shpg'),
    SIZE=16
    )

shader_index = Struct("shader_index",
    dyn_senum32("shader",
        DYN_NAME_PATH="tagdata.efpg_attrs.shaders.STEPTREE[DYN_I].shader.filepath"),
    SIZE=4
    )

exposed_parameter = Struct("exposed_parameter",
    ascii_str32("exposed_name"),
    ascii_str32("parameter_name"),
    dyn_senum32("shader_index",
        DYN_NAME_PATH="tagdata.efpg_attrs.shader_indices.STEPTREE[DYN_I].NAME"),
    SIZE=80
    )

efpg_attrs = Struct("efpg_attrs",
    reflexive("shaders", shader, 12,
        DYN_NAME_PATH='.shader.filepath'),
    reflexive("shader_indices", shader_index, 12),
    reflexive("exposed_parameters", exposed_parameter, 32,
        DYN_NAME_PATH='.exposed_name'),
    SIZE=36
    )

efpg_body = Struct("tagdata",
    efpp_attrs,
    efpg_attrs,
    SIZE=96
    )

def get():
    return efpg_def

efpg_def = TagDef("efpg",
    blam_header_os('efpg'),
    efpg_body,

    ext=".effect_postprocess_generic", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from .shdr import *
from .objs.shdr import ShdrTag
from supyr_struct.defs.tag_def import TagDef

environment_shader_comment = """ENVIRONMENT SHADER
Setting <true atmospheric fog> enables per-pixel atmospheric fog (for models) but disables
point/spot lights, planar fog, and the ability to control the atmospheric fog density for
this shader.

Alpha-tested controls if the shader is masked by the diffuse alpha (or depending on
the environment shader type's setting) the bump map alpha. The transparancy is only 1-bit"""

environment_shader_type_comment = """ENVIRONMENT SHADER TYPE
Controls how diffuse maps are combined:

NORMAL:
Secondary detail map alpha controls blend between primary and secondary detail map.
Specular mask is alpha of blended primary and secondary detail map alpha multiplied by
alpha of micro detail map.

BLENDED:
Base map alpha controls blend between primary and secondary detail map.
Specular mask is alpha of blended primary and secondary detail map alpha multiplied by
alpha of micro detail map.

BLENDED BASE SPECULAR:
Same as BLENDED, except specular mask is alpha is base map multiplied with
the alpha of micro detail map."""

bump_properties_comment = """BUMP PROPERTIES
Perforated (alpha-tested) shaders use alpha in bump map."""

tex_scroll_anim_comment = """TEXTURE SCROLLING ANIMATION
Scrolls all 2D maps simultaneously."""

self_illum_comment = """SELF-ILLUMINATION PROPERTIES
There are three self-illumination effects which are added together.
Each effect has an <on color>, used when the shader is active, and an <off color>, used when
the shader is not active. The self-illumination map is used as follows:
* RED: primary mask
* GREEN: secondary mask
* BLUE: plasma mask
* ALPHA: plasma animation reference

Each effect also has an animation <function>, <period> and <phase>, used when the shader is
active. The primary and secondary effects simply modulate the <on color> by the animation
value to produce an animation color, and then blend between the animation color and the
<off color> based on the shader's activation level, and finally modulate by the mask.

The plasma shader compares the animation value with the alpha channel of the map (the plasma
animation reference) and produces a high value when they are similar and a dark value when
they are different. This value modulates the <plasma on color> to produce a plasma animation
color, and the rest proceeds just like the primary and secondary effects."""

specular_properties_comment = """SPECULAR PROPERTIES
Controls the dynamic specular highlights. The highlight is modulated by brightness
and a blend between perpendicular and parrallel color.

These color values also affect the colour of the reflection in REFLECTION PROPERTIES"""

reflection_properties_comment = """REFLECTION PROPERTIES
Controls cube map reflections. The color of the cubemap is tinted by the color settings
in the SPECULAR PROPERTIES and the brightness in the REFLECTION PROPERTIES.

BUMPED CUBE MAP: Makes it so that the reflection and fresnel is affected by the bump map.

FLAT CUBE MAP: The reflection is not affected by the cubemap, the fresnel still is though."""

environment_shader = Struct("environment_shader",
    Bool16("flags",
        "alpha_tested",
        "bump_map_is_specular_mask",
        "true_atmospheric_fog",
        COMMENT=environment_shader_comment
        ),
    SEnum16("type",
        "normal",
        "blended",
        "blended_base_specular",
        COMMENT=environment_shader_type_comment
        ),
    )

diffuse = Struct("diffuse",
    Bool16("diffuse_flags",
        "rescale_detail_maps",
        "rescale_bump_maps",
        ),
    Pad(26),
    dependency("base_map", "bitm"),

    Pad(24),
    SEnum16("detail_map_function", *detail_map_functions),
    Pad(2),

    Float("primary_detail_map_scale"),
    dependency("primary_detail_map", "bitm"),
    Float("secondary_detail_map_scale"),
    dependency("secondary_detail_map", "bitm"),

    Pad(24),
    SEnum16("micro_detail_map_function", *detail_map_functions),

    Pad(2),
    Float("micro_detail_map_scale"),
    dependency("micro_detail_map", "bitm"),
    QStruct("material_color", INCLUDE=rgb_float),
    )

bump_properties = Struct("bump_properties",
    Float("map_scale"),
    dependency("map", "bitm"),
    FlFloat("map_scale_x", VISIBLE=False),
    FlFloat("map_scale_y", VISIBLE=False),
    COMMENT=bump_properties_comment
    )

texture_scrolling = Struct("texture_scrolling",
    anim_func_per_sca_macro("u_animation"),
    anim_func_per_sca_macro("v_animation"),
    COMMENT=tex_scroll_anim_comment
    )

self_illumination = Struct("self_illumination",
    Bool16("flags",
        "unfiltered",
        ),
    Pad(2),
    Pad(24),

    QStruct("primary_on_color",  INCLUDE=rgb_float),
    QStruct("primary_off_color", INCLUDE=rgb_float),
    anim_func_per_pha_macro("primary_animation"),

    Pad(24),
    QStruct("secondary_on_color",  INCLUDE=rgb_float),
    QStruct("secondary_off_color", INCLUDE=rgb_float),
    anim_func_per_pha_macro("secondary_animation"),

    Pad(24),
    QStruct("plasma_on_color",  INCLUDE=rgb_float),
    QStruct("plasma_off_color", INCLUDE=rgb_float),
    anim_func_per_pha_macro("plasma_animation"),

    Pad(24),
    Float("map_scale"),
    dependency("map", "bitm"),
    COMMENT=self_illum_comment
    )

specular = Struct("specular",
    Bool16("specular_flags",
        "overbright",
        "extra_shiny",
        "lightmap_is_specular"
        ),
    Pad(18),
    float_zero_to_one("brightness"),  # [0,1]

    Pad(20),
    QStruct("perpendicular_tint_color", INCLUDE=rgb_float),
    QStruct("parallel_tint_color",      INCLUDE=rgb_float),
    COMMENT=specular_properties_comment
    )

reflection = Struct("reflection",
    Bool16("reflection_flags",
        "dynamic_mirror",
         ),
    SEnum16("reflection_type",
        "bumped_cubemap",
        "flat_cubemap",
        "bumped_radiosity",
        ),

    float_zero_to_one("lightmap_brightness_scale"),  # [0,1]
    Pad(28),
    float_zero_to_one("perpendicular_brightness"),  # [0,1]
    float_zero_to_one("parallel_brightness"),  # [0,1]

    Pad(40),
    dependency("cube_map", "bitm"),
    COMMENT=reflection_properties_comment
    )

senv_attrs = Struct("senv_attrs",
    #Environment Shader Properties
    environment_shader,

    #Lens Flare
    float_wu("lens_flare_spacing"),  # world units
    dependency("lens_flare", "lens"),

    # this padding is the reflexive for the OS shader environment extension
    Pad(12),

    Pad(32),
    #Diffuse Properties
    diffuse,

    Pad(12),
    #Bump Properties
    bump_properties,

    Pad(16),
    #Texture Animation
    texture_scrolling,

    Pad(24),
    #Self Illumination
    self_illumination,

    Pad(24),
    #Specular Properties
    specular,

    Pad(16),
    #Reflection Properties
    reflection,
    SIZE=796
    )

senv_body = Struct("tagdata",
    shdr_attrs,
    senv_attrs,
    SIZE=836,
    )


def get():
    return senv_def

senv_def = TagDef("senv",
    blam_header('senv', 2),
    senv_body,

    ext=".shader_environment", endian=">", tag_cls=ShdrTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Deadcanadian
# 	naming stuff
# revision: 3		author: OrangeMohawk
# 	Data types
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


dsrc_data_integer_value = Struct("integer_value",
    h3_string_id("data_type"),
    SInt32("value"),
    ENDIAN=">", SIZE=8
    )


dsrc_data_string_value = Struct("string_value",
    h3_string_id("data_type"),
    StrLatin1("value", SIZE=20),
    ENDIAN=">", SIZE=36
    )


dsrc_data_stringid_value = Struct("stringid_value",
    h3_string_id("data_type"),
    h3_string_id("value"),
    ENDIAN=">", SIZE=8
    )


dsrc_data = Struct("data",
    h3_reflexive("integer_values", dsrc_data_integer_value),
    h3_reflexive("string_values", dsrc_data_string_value),
    h3_reflexive("stringid_values", dsrc_data_stringid_value),
    h3_string_id("unknown"),
    ENDIAN=">", SIZE=40
    )


dsrc_body = Struct("tagdata",
    h3_string_id("name"),
    BytesRaw("unknown", SIZE=12, VISIBLE=False),
    h3_reflexive("data", dsrc_data),
    ENDIAN=">", SIZE=28
    )


def get():
    return dsrc_def

dsrc_def = TagDef("dsrc",
    h3_blam_header('dsrc'),
    dsrc_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["dsrc"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

def get():
    return buym_def

buym_def = TagDef("buym",
    blam_header('buym'),
    Struct('tagdata',
        dependency("forward", "snd!"),
        dependency("back", "snd!"),
        dependency("click_1", "snd!"),
        dependency("click_2", "snd!"),
        dependency("click_3", "snd!"),
        dependency("click_4", "snd!"),
        dependency("sound_7", "snd!"),
        dependency("cursor", "snd!"),
        dependency("sound_9", "snd!"),
        dependency("sound_10", "snd!"),
        dependency("sound_11", "snd!"),
        dependency("sound_12", "snd!"),
        dependency("sound_13", "snd!"),
        dependency("sound_14", "snd!"),
        dependency("sound_15", "snd!"),
        dependency("sound_16", "snd!"),
        dependency("sound_17", "snd!"),
        dependency("sound_18", "snd!"),
        dependency("sound_19", "snd!"),
        dependency("sound_20", "snd!"),
        dependency("select_magic", "snd!"),
        dependency("select_tech", "snd!"),
        dependency("assign_magic", "snd!"),
        dependency("assign_tech", "snd!"),
        dependency("sound_25", "snd!"),

        Pad(1344),

        dependency("sound_26", "snd!"),
        dependency("sound_27", "snd!"),
        dependency("sound_28", "snd!"),
        dependency("sound_29", "snd!"),
        SIZE=2000,
        ),

    ext=".buy_menu", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.sbsp import *



# ISC License (ISC)
#
# Copyright 2017 Devin Bobadilla
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import array
import audioop
import sys

from struct import unpack_from
from types import MethodType

from . import constants

try:
    from .ext import adpcm_ext
    fast_adpcm = True
except:
    fast_adpcm = False

__all__ = (
    "decode_adpcm_samples",
    "encode_adpcm_samples",
    "NOISE_SHAPING_OFF", "NOISE_SHAPING_STATIC", "NOISE_SHAPING_DYNAMIC",
    )

NIBBLE_SWAP_MAPPING = tuple(((val&15)<<4) | (val>>4) for val in range(256))


NOISE_SHAPING_OFF     = 0  # flat noise (no shaping)
NOISE_SHAPING_STATIC  = 1  # first-order highpass shaping
NOISE_SHAPING_DYNAMIC = 2  # dynamically tilted noise based on signal


def _slow_decode_xbadpcm_samples(in_data, out_data, channel_ct):
    # divide by 2 since we're treating out_data as a sint16 iterable
    pcm_blocksize   = channel_ct * constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2
    adpcm_blocksize = channel_ct * constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE

    adpcm2lin = audioop.adpcm2lin
    all_codes = memoryview(in_data)

    state_unpacker = MethodType(unpack_from, "<" + "hBx" * channel_ct)
    code_block_size = 4 * channel_ct

    k = 0
    interleaved = channel_ct > 1
    for i in range(0, len(in_data), adpcm_blocksize):
        state_vals = state_unpacker(in_data, i)
        all_states = tuple(zip(state_vals[::2], state_vals[1::2]))
        all_swapped_codes = bytes(map(
            NIBBLE_SWAP_MAPPING.__getitem__,
            all_codes[i + code_block_size: i + adpcm_blocksize]))

        for c in range(channel_ct):
            # join all 4 bytes codes for this channel
            swapped_codes = all_swapped_codes
            if interleaved:
                swapped_codes = b''.join(
                    # join the 4 byte codes
                    swapped_codes[j: j + 4]
                    for j in range(c * 4, len(swapped_codes), code_block_size)
                    )
            decoded_samples = memoryview(
                adpcm2lin(swapped_codes, 2, all_states[c])[0]).cast("h")

            # interleave the samples for each channel
            out_data[k + c] = all_states[c][0]
            out_data[k + c + channel_ct: k + pcm_blocksize: channel_ct] = array.array(
                "h", decoded_samples)[: -1]

        k += pcm_blocksize


def decode_adpcm_samples(in_data, channel_ct, output_big_endian=False):
    if channel_ct < 1:
        return b''

    # divide by 2 since we're treating out_data as a sint16 iterable
    out_data = array.array("h", (0,)) * (
        (len(in_data) // constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE) *
        (constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE // 2))

    if fast_adpcm:
        adpcm_ext.decode_xbadpcm_samples(in_data, out_data, channel_ct)
    else:
        _slow_decode_xbadpcm_samples(in_data, out_data, channel_ct)

    if (sys.byteorder == "big") != output_big_endian:
        out_data.byteswap()

    return bytes(out_data)


def encode_adpcm_samples(in_data, channel_ct, input_big_endian=False,
                         noise_shaping=NOISE_SHAPING_OFF, lookahead=3):
    assert noise_shaping in (NOISE_SHAPING_OFF, NOISE_SHAPING_STATIC, NOISE_SHAPING_DYNAMIC)
    assert lookahead in range(6)

    if channel_ct < 1:
        return b''
    elif not fast_adpcm:
        raise NotImplementedError(
            "Accelerator module not detected. Cannot compress to ADPCM.")

    if (sys.byteorder == "big") != input_big_endian:
        out_data = audioop.byteswap(out_data, 2)

    adpcm_blocksize = constants.XBOX_ADPCM_COMPRESSED_BLOCKSIZE * channel_ct
    pcm_blocksize = constants.XBOX_ADPCM_DECOMPRESSED_BLOCKSIZE * channel_ct

    pad_size = len(in_data) % pcm_blocksize
    if pad_size:
        pad_size = pcm_blocksize - pad_size
        # repeat the last sample to the end to pad to a multiple of blocksize
        pad_piece_size = (channel_ct * 2)
        in_data += in_data[-pad_piece_size: ] * (pad_size // pad_piece_size)

    out_data = bytearray(
        (len(in_data) // pcm_blocksize) * adpcm_blocksize
        )

    adpcm_ext.encode_xbadpcm_samples(
        in_data, out_data, channel_ct, noise_shaping, lookahead)

    return bytes(out_data)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sqrt
from struct import Struct as PyStruct

from reclaimer.model.constants import (
    HALO_1_MAX_MATERIALS, HALO_1_MAX_REGIONS, HALO_1_MAX_GEOMETRIES_PER_MODEL,
    SCALE_INTERNAL_TO_JMS, HALO_1_NAME_MAX_LEN
    )
from reclaimer.model.jms import GeometryMesh
from reclaimer.model.stripify import Stripifier
from reclaimer.model import util


def compile_gbxmodel(mod2_tag, merged_jms, ignore_errors=False):
    tagdata = mod2_tag.data.tagdata

    tagdata.flags.parts_have_local_nodes = False

    u_scale, v_scale = merged_jms.calc_uv_scales()
    if u_scale < 1:
        u_scale = 1
    if v_scale < 1:
        v_scale = 1
    tagdata.base_map_u_scale = merged_jms.u_scale = u_scale
    tagdata.base_map_v_scale = merged_jms.v_scale = v_scale

    tagdata.node_list_checksum = merged_jms.node_list_checksum


    errors = []
    if len(merged_jms.materials) > HALO_1_MAX_MATERIALS:
        errors.append("Too many materials. Max count is %s." % (HALO_1_MAX_MATERIALS))

    if len(merged_jms.regions) > HALO_1_MAX_REGIONS:
        errors.append("Too many regions. Max count is %s." % (HALO_1_MAX_REGIONS))

    if errors and not ignore_errors:
        return errors

    # make nodes
    mod2_nodes = tagdata.nodes.STEPTREE
    del mod2_nodes[:]
    for node in merged_jms.nodes:
        mod2_nodes.append()
        mod2_node = mod2_nodes[-1]

        mod2_node.name = node.name[: HALO_1_NAME_MAX_LEN]
        mod2_node.next_sibling_node = node.sibling_index
        mod2_node.first_child_node = node.first_child
        mod2_node.parent_node = node.parent_index
        mod2_node.translation[:] = node.pos_x / SCALE_INTERNAL_TO_JMS,\
                                   node.pos_y / SCALE_INTERNAL_TO_JMS,\
                                   node.pos_z / SCALE_INTERNAL_TO_JMS
        mod2_node.rotation[:] = node.rot_i, node.rot_j,\
                                node.rot_k, node.rot_w

        if node.parent_index >= 0:
            mod2_node.distance_from_parent = sqrt(
                node.pos_x**2 + node.pos_y**2 + node.pos_z**2) / SCALE_INTERNAL_TO_JMS


    # record shader ordering and permutation indices
    mod2_shaders = tagdata.shaders.STEPTREE
    shdr_perm_indices_by_name = {}
    for mod2_shader in mod2_shaders:
        shdr_name = mod2_shader.shader.filepath.split("\\")[-1]
        shdr_perm_indices = shdr_perm_indices_by_name.setdefault(shdr_name, [])
        shdr_perm_indices.append(mod2_shader.permutation_index)

    del mod2_shaders[:]
    # make shader references
    for mat in merged_jms.materials:
        mod2_shaders.append()
        mod2_shader = mod2_shaders[-1]
        mod2_shader.shader.filepath = mat.shader_path
        if mat.shader_type:
            mod2_shader.shader.tag_class.set_to(mat.shader_type)
        else:
            mod2_shader.shader.tag_class.set_to("shader")

        shdr_name = mod2_shader.shader.filepath.split("\\")[-1].lower()
        shdr_perm_indices = shdr_perm_indices_by_name.get(shdr_name)
        if shdr_perm_indices:
            mod2_shader.permutation_index = shdr_perm_indices.pop(0)


    # make regions
    mod2_regions = tagdata.regions.STEPTREE
    del mod2_regions[:]

    global_markers = {}
    geom_meshes = []
    all_lod_nodes = {lod: set([0]) for lod in util.LOD_NAMES}
    for region_name in sorted(merged_jms.regions):
        region = merged_jms.regions[region_name]

        mod2_regions.append()
        mod2_region = mod2_regions[-1]
        mod2_region.name = region_name[: HALO_1_NAME_MAX_LEN]

        mod2_perms = mod2_region.permutations.STEPTREE
        for perm_name in sorted(region.perm_meshes):
            perm = region.perm_meshes[perm_name]

            mod2_perms.append()
            mod2_perm = mod2_perms[-1]
            mod2_perm.name = perm_name[: HALO_1_NAME_MAX_LEN]

            mod2_perm.flags.cannot_be_chosen_randomly = not perm.is_random_perm

            perm_added = False
            skipped_lods = []
            for i in range(len(util.LOD_NAMES)):
                lod_name = util.LOD_NAMES[i]
                if not perm.lod_meshes.get(lod_name):
                    if skipped_lods is not None:
                        skipped_lods.append(i)
                    continue

                geom_index = len(geom_meshes)
                lod_mesh = perm.lod_meshes[lod_name]
                geom_meshes.append(lod_mesh)

                # figure out which nodes this mesh utilizes
                this_meshes_nodes = set()
                for mesh in lod_mesh.values():
                    for vert in mesh.verts:
                        if vert.node_1_weight < 1:
                            this_meshes_nodes.add(vert.node_0)
                        if vert.node_1_weight > 0:
                            this_meshes_nodes.add(vert.node_1)

                all_lod_nodes[lod_name].update(this_meshes_nodes)

                lods_to_set = list(range(i, 5))
                if skipped_lods:
                    lods_to_set.extend(skipped_lods)
                    skipped_lods = None

                for i in lods_to_set:
                    setattr(mod2_perm,
                            "%s_geometry_block" % util.LOD_NAMES[i],
                            geom_index)

                perm_added = True

            # What are we doing here?
            if len(perm.markers) > 32:
                for marker in perm.markers:
                    global_markers.setdefault(
                        marker.name[: HALO_1_NAME_MAX_LEN], []).append(marker)
            else:
                perm_added |= bool(perm.markers)
                mod2_markers = mod2_perm.local_markers.STEPTREE
                for marker in perm.markers:
                    mod2_markers.append()
                    mod2_marker = mod2_markers[-1]

                    mod2_marker.name = marker.name[: HALO_1_NAME_MAX_LEN]
                    mod2_marker.node_index = marker.parent
                    mod2_marker.translation[:] = marker.pos_x / SCALE_INTERNAL_TO_JMS,\
                                                 marker.pos_y / SCALE_INTERNAL_TO_JMS,\
                                                 marker.pos_z / SCALE_INTERNAL_TO_JMS
                    mod2_marker.rotation[:] = marker.rot_i, marker.rot_j,\
                                              marker.rot_k, marker.rot_w


            if not(perm_added or ignore_errors):
                del mod2_perms[-1]
                continue

    if len(geom_meshes) > HALO_1_MAX_GEOMETRIES_PER_MODEL and not ignore_errors:
        return ("Cannot add more than %s geometries to a model. "
                "Each material in each region in each permutation "
                "in each LOD is counted as a single geometry.\n"
                "This model would contain %s geometries." % (
                    HALO_1_MAX_GEOMETRIES_PER_MODEL, len(geom_meshes)), )

    # make the markers
    mod2_marker_headers = tagdata.markers.STEPTREE
    del mod2_marker_headers[:]
    for marker_name in sorted(global_markers):
        marker_list = global_markers[marker_name]
        mod2_marker_headers.append()
        mod2_marker_header = mod2_marker_headers[-1]

        mod2_marker_header.name = marker_name[: HALO_1_NAME_MAX_LEN]
        mod2_marker_list = mod2_marker_header.marker_instances.STEPTREE

        for marker in marker_list:
            mod2_marker_list.append()
            mod2_marker = mod2_marker_list[-1]

            # figure out which permutation index this marker
            # matches for all the permutations in its region
            i = perm_index = 0
            for perm in mod2_regions[marker.region].permutations.STEPTREE:
                if perm.name == marker.permutation:
                    perm_index = i
                    break
                i += 1

            mod2_marker.region_index = marker.region
            mod2_marker.permutation_index = perm_index
            mod2_marker.node_index = marker.parent
            mod2_marker.translation[:] = marker.pos_x / SCALE_INTERNAL_TO_JMS,\
                                         marker.pos_y / SCALE_INTERNAL_TO_JMS,\
                                         marker.pos_z / SCALE_INTERNAL_TO_JMS
            mod2_marker.rotation[:] = marker.rot_i, marker.rot_j,\
                                      marker.rot_k, marker.rot_w

    # set the node counts per lod
    for lod in util.LOD_NAMES:
        lod_nodes = all_lod_nodes[lod]
        adding = True
        node_ct = len(mod2_nodes)

        for i in range(node_ct - 1, -1, -1):
            if i in lod_nodes:
                break
            node_ct -= 1

        setattr(tagdata, "%s_lod_nodes" % lod, max(0, node_ct - 1))


    # calculate triangle strips
    stripped_geom_meshes = []
    for geom_idx in range(len(geom_meshes)):
        material_meshes = {}
        stripped_geom_meshes.append(material_meshes)
        for mat_idx in sorted(geom_meshes[geom_idx]):
            material_meshes[mat_idx] = mesh_list = []
            geom_mesh = geom_meshes[geom_idx][mat_idx]
            all_verts = geom_mesh.verts

            stripifier = Stripifier()
            stripifier.load_mesh(geom_mesh.tris, True)
            stripifier.make_strips()
            stripifier.link_strips()

            if stripifier.get_strip_count() == 1:
                tri_strip = stripifier.translate_strip(stripifier.get_strip())
            else:
                all_verts = util.EMPTY_GEOM_VERTS
                tri_strip = (0, 1, 2)

            if len(tri_strip) > util.MAX_STRIP_LEN:
                return (
                    ("Too many triangles ya fuck. Max triangles per "
                     "geometry is %s.\nThis geometry is %s after linking "
                     "all strips.") % (util.MAX_STRIP_LEN, len(tri_strip)), )

            mesh_list.append(GeometryMesh(all_verts, tri_strip))


    # make the geometries
    mod2_geoms = tagdata.geometries.STEPTREE
    del mod2_geoms[:]
    vert_packer = PyStruct(">14f2h2f").pack_into
    for geom_idx in range(len(stripped_geom_meshes)):
        mod2_geoms.append()
        mod2_parts = mod2_geoms[-1].parts.STEPTREE

        for mat_idx in sorted(stripped_geom_meshes[geom_idx]):
            geom_mesh_list = stripped_geom_meshes[geom_idx][mat_idx]
            for geom_mesh in geom_mesh_list:
                mod2_parts.append()
                mod2_part = mod2_parts[-1]
                mod2_verts = mod2_part.uncompressed_vertices.STEPTREE

                tris  = geom_mesh.tris
                verts = geom_mesh.verts
                vert_ct = len(verts)
                mod2_verts.extend(len(verts))
                mod2_part.shader_index = mat_idx

                cent_x = cent_y = cent_z = 0

                # TODO: Modify this to take into account local nodes
                # honestly though, who the fuck is going to care? fuck it.

                # make a raw vert reflexive and replace the one in the part
                mod2_part.uncompressed_vertices = util.mod2_verts_def.build()
                mod2_verts = mod2_part.uncompressed_vertices.STEPTREE = \
                             bytearray(68 * len(verts))
                i = 0
                for vert in verts:
                    vert_packer(
                        mod2_verts, i,
                        vert.pos_x / SCALE_INTERNAL_TO_JMS,
                        vert.pos_y / SCALE_INTERNAL_TO_JMS,
                        vert.pos_z / SCALE_INTERNAL_TO_JMS,
                        vert.norm_i, vert.norm_j, vert.norm_k,
                        vert.binorm_i, vert.binorm_j, vert.binorm_k,
                        vert.tangent_i, vert.tangent_j, vert.tangent_k,
                        vert.tex_u / u_scale, (1.0 - vert.tex_v) / v_scale,
                        vert.node_0, vert.node_1,
                        1 - vert.node_1_weight, vert.node_1_weight)
                    i += 68
                    cent_x += vert.pos_x / (vert_ct * SCALE_INTERNAL_TO_JMS)
                    cent_y += vert.pos_y / (vert_ct * SCALE_INTERNAL_TO_JMS)
                    cent_z += vert.pos_z / (vert_ct * SCALE_INTERNAL_TO_JMS)

                mod2_part.centroid_translation[:] = [cent_x, cent_y, cent_z]

                # make a raw tri reflexive and replace the one in the part
                mod2_part.triangles = util.mod2_tri_strip_def.build()
                mod2_tris = mod2_part.triangles.STEPTREE = bytearray(
                    [255, 255]) * (3 * ((len(tris) + 2) // 3))
                i = 0
                for tri in tris:
                    mod2_tris[i]     = tri >> 8
                    mod2_tris[i + 1] = tri & 0xFF
                    i += 2



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import ceil, log

from arbytmap.bitmap_io import get_pixel_bytes_size
from reclaimer.h3.constants import HALO3_SHARED_MAP_TYPES
from reclaimer.util import *


def get_virtual_dimension(bitm_fmt, dim, mip_level=0, tiled=False):
    dim = max(1, dim >> mip_level)
    if bitm_fmt in ("A8R8G8B8", "X8R8G8B8"):
        stride = 32
    elif bitm_fmt in ("A8", "L8", "AL8", "A8L8", "V8U8",
                      "R5G6B5", "A1R5G5B5", "A4R4G4B4"):
        stride = 64
    else:
        stride = 128

    dim += ((stride - (dim % stride)) % stride)
    if mip_level != 0 and tiled:
        # first bitmap isnt padded
        dim = 2**int(ceil(log(dim, 2.0)))

    return dim


def get_h3_pixel_bytes_size(bitm_fmt, width, height, depth, mip=0, tiled=False):
    #if width <= 16 or height <= 16:
    #    tiled = False

    width = get_virtual_dimension(bitm_fmt, width, mip, tiled)
    height = get_virtual_dimension(bitm_fmt, height, mip, tiled)
    return get_pixel_bytes_size(bitm_fmt, width, height, depth)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.sotr import *
from .shdr import *
from supyr_struct.defs.tag_def import TagDef

sotr_body = Struct("tagdata",
    shdr_attrs,
    sotr_attrs,

    SIZE=108,
    )

def get():
    return sotr_def

sotr_def = TagDef("sotr",
    blam_header_stubbs("sotr"),
    sotr_body,

    ext=".shader_transparent_generic", endian=">", tag_cls=ShdrTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.schi import *
from .shdr import *
from supyr_struct.defs.tag_def import TagDef

schi_body = Struct("tagdata",
    shdr_attrs,
    schi_attrs,
    SIZE=108
    )

def get():
    return schi_def

schi_def = TagDef("schi",
    blam_header_stubbs('schi'),
    schi_body,

    ext=".shader_transparent_chicago", endian=">", tag_cls=ShdrTag,
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Mapped out most of the tag
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


rasg_default_bitmap = Struct("default_bitmap",
    SInt32("unknown", VISIBLE=False),
    h3_dependency("bitmap"),
    ENDIAN=">", SIZE=20
    )


rasg_default_rasterizer_bitmap = Struct("default_rasterizer_bitmap",
    h3_dependency("bitmap"),
    ENDIAN=">", SIZE=16
    )


rasg_default_shader = Struct("default_shader",
    h3_dependency("vertex_shader"),
    h3_dependency("pixel_shader"),
    ENDIAN=">", SIZE=32
    )


rasg_body = Struct("tagdata",
    h3_reflexive("default_bitmaps", rasg_default_bitmap),
    h3_reflexive("default_rasterizer_bitmaps", rasg_default_rasterizer_bitmap),
    h3_dependency("vertex_shader_simple"),
    h3_dependency("pixel_shader_simple"),
    h3_reflexive("default_shaders", rasg_default_shader),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SInt32("unknown_1", VISIBLE=False),
    SInt32("unknown_2", VISIBLE=False),
    h3_dependency("active_camo_distortion"),
    h3_dependency("default_performance_template"),
    h3_dependency("default_shield_impact"),
    SInt32("unknown_3", VISIBLE=False),
    BytesRaw("unknown_4", SIZE=24, VISIBLE=False),
    ENDIAN=">", SIZE=164
    )


def get():
    return rasg_def

rasg_def = TagDef("rasg",
    h3_blam_header('rasg'),
    rasg_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["rasg"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.obje import ObjeTag

class WeapTag(ObjeTag):

    def calc_internal_data(self):
        ObjeTag.calc_internal_data(self)
        for trigger in self.data.tagdata.weap_attrs.triggers.STEPTREE:
            firing = trigger.firing
            misc = trigger.misc
            rates = trigger.misc_rates
            for i in range(len(rates)):
                rates[i] = 0

            if misc.ejection_port_recovery_time:
                rates.ejection_port_recovery_rate = 1 / misc.ejection_port_recovery_time
            if misc.illumination_recovery_time:
                rates.illumination_recovery_rate = 1 / misc.illumination_recovery_time

            if firing.acceleration_time:
                rates.acceleration_rate = 1 / firing.acceleration_time
            if firing.deceleration_time:
                rates.deceleration_rate = 1 / firing.deceleration_time

            if firing.error_acceleration_time:
                rates.error_acceleration_rate = 1 / firing.error_acceleration_time
            if firing.error_deceleration_time:
                rates.error_deceleration_rate = 1 / firing.error_deceleration_time

            for i in range(len(rates)):
                rates[i] /= 30



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

material = Struct("material",
    dependency_stubbs("effect", "effe"),
    dependency_stubbs("sound", "snd!"),
    SIZE=48,
    )

effect = Struct("effect",
    reflexive("materials", material, len(materials_list), *materials_list),
    SIZE=28,
    )

foot_body = Struct("tagdata",
    reflexive("effects", effect, 13,
        "walk", "run", "sliding", "shuffle", "jump", "jump land",
        "biped unused1", "biped unused2",
        "impact", "vehicle tire slip", "vehicle chassis slip",
        "vehicle unused1", "vehicle unused2"),
    SIZE=140,
    )



def get():
    return foot_def

foot_def = TagDef("foot",
    blam_header_stubbs('foot'),
    foot_body,

    ext=".material_effects", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: -DeToX-
# 	Created layout of plugin
# revision: 2		author: Lord Zedd
# 	Done
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


chmt_lighting_variable = Struct("lighting_variable",
    Float("lightmap_brightness_offset"),
    ENDIAN=">", SIZE=4
    )


chmt_body = Struct("tagdata",
    h3_reflexive("lighting_variables", chmt_lighting_variable),
    ENDIAN=">", SIZE=12
    )


def get():
    return chmt_def

chmt_def = TagDef("chmt",
    h3_blam_header('chmt'),
    chmt_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["chmt"], endian=">", tag_cls=H3Tag
    )
